#!/usr/bin/perl

use strict;
use warnings;
use Data::Dumper;
use File::Temp qw/ tempfile tempdir /;

# Output file of cmsearch in tabular format
my $file_in = "/home/tanja/holo_genomes_results/Lmig_cmsearch_g.tbl";
open my $fh_in, '<', $file_in or die "Couldn't open file \"$file_in\": $!\n";

# File with non eukaryotic Rfam families
my $file_rfam = "/home/tanja/Rfam_12/non_eukaryotes_nc.1st_column";
open my $fh_rfam, '<', $file_rfam or die "Couldn't open file \"$file_rfam\": $!\n";

my %non_eu_rfam;

while ( my $line = <$fh_rfam> ) {
	chomp $line;
	$line =~ m/(RF\d{5}).*/;
	$non_eu_rfam{ $1 } += 1;
}

my %hit_of_scaff;

# Get all hits from the cmsearch out file where e-vaule <= 0.01
while ( my $line = <$fh_in> ) {
	chomp $line;
	next if $line =~ m/^#/;
	my @parts = split ( ' ', $line );
	if ( $parts[ 16 ] =~ m/!/ and !exists $non_eu_rfam{ $parts[ 3 ] } ) {
		push @{ $hit_of_scaff{ $parts[ 0 ]} }, { 'start' => $parts[ 7 ], 'end' => $parts[ 8 ], 'hit' => $line, 'id' => $parts[ 3 ] };
	}
}

close $fh_in;

# File to save hits according to criteria and only once

my $file_final = "/home/tanja/holo_genomes_results/Lmig_cmsearch_g_sorted.tbl";
open my $fh_final, '>', $file_final or die "Couldn't open file \"$file_final\": $!\n";

# File to save hits which have the exact same coordinates
my $file_duplicates = "/home/tanja/holo_genomes_results/Lmig_cmsearch_g_duplicates";
open my $fh_dupl, '>', $file_duplicates or die "Couldn't open file \"$file_duplicates\": $!\n";

# Count non redundant number of Rfam families (hits save in fh_dupl will be only saved once)
my $file_id_count = "/home/tanja/holo_genomes_results/Lmig_cmsearch_g_id_count";
open my $fh_id, '>', $file_id_count or die "Couldn't open file \"$file_id_count\": $!\n";

# File to save hits that overlap
my $file_overlap = "/home/tanja/holo_genomes_results/Lmig_cmsearch_g_overlap";
open my $fh_ol, '>', $file_overlap or die "Couldn't open file \"$file_overlap\": $!\n";

my %count_of;

# Iterate through each scaffold
foreach my $key ( sort keys %hit_of_scaff ) {
	my $start;
	my $end;
	
	# Iterate through each hit in scaffold and get the start and stop coordinates of the hit
	for ( my $i = 0; $i <= $#{ $hit_of_scaff{ $key } }; ++$i ){
		if ( $hit_of_scaff{ $key }[ $i ]{ 'start' } < $hit_of_scaff{ $key }[ $i ]{ 'end' } ){
			$start = $hit_of_scaff{ $key }[ $i ]{ 'start' };
			$end   = $hit_of_scaff{ $key }[ $i ]{ 'end' };
		}
		else {
			$start = $hit_of_scaff{ $key }[ $i ]{ 'end' };
			$end   = $hit_of_scaff{ $key }[ $i ]{ 'start' };
		}
		my $start2;
		my $end2;
		my $first = 1;
		my $first_ol = 1;
		
		# Compare the coordinates of the hit with the following hits
		for ( my $j = $i + 1; $j <= $#{ $hit_of_scaff{ $key } }; ++$j ){
			if ( $hit_of_scaff{ $key }[ $j ]{ 'start' } < $hit_of_scaff{ $key }[ $j ]{ 'end' } ){
				$start2 = $hit_of_scaff{ $key }[ $j ]{ 'start' };
				$end2   = $hit_of_scaff{ $key }[ $j ]{ 'end' };
			}
			else {
				$start2 = $hit_of_scaff{ $key }[ $j ]{ 'end' };
				$end2   = $hit_of_scaff{ $key }[ $j ]{ 'start' };
			}
			
			# Check if both hits have the same coordinates
			if ( $start == $start2 and $end == $end2 ) {
				# Print hits to fh_dupl, count it only once in count_of
				if ( $first ) {
					print { $fh_dupl } $hit_of_scaff{ $key }[ $i ]{ 'hit' }, "\n";
					$first = 0;
					$count_of{ $hit_of_scaff{ $key }[ $i ]{ 'id' } } += 1;
				}
				print { $fh_dupl } $hit_of_scaff{ $key }[ $j ]{ 'hit' }, "\n";
			}
			
			# Check if both hits overlap 
			
			# noch eine Flag einbauen, damit duplicates nur 1x gezählt werden
			elsif ( $start < $start2 and $start <= $end2 and $end >= $start2 ) {
				if ( $first_ol ) {
					print { $fh_ol } $hit_of_scaff{ $key }[ $i ]{ 'hit' }, "\n";
					print { $fh_final } $hit_of_scaff{ $key }[ $i ]{ 'hit' }, "\n";
					$first_ol = 0;
				}
				print { $fh_ol } $hit_of_scaff{ $key }[ $j ]{ 'hit' }, "\n";
			}
			elsif ( $start > $start2 and $start <= $end2 and $end >= $start2 ) {
				if ( $first_ol ) {
					print { $fh_ol } $hit_of_scaff{ $key }[ $i ]{ 'hit' }, "\n";
					print { $fh_final } $hit_of_scaff{ $key }[ $i ]{ 'hit' }, "\n";
					$first_ol = 0;
				}
				print { $fh_ol } $hit_of_scaff{ $key }[ $j ]{ 'hit' }, "\n";
			}
			elsif ( $start >= $start2 and $start < $end2 ) {
				if ( $first_ol ) {
					print { $fh_ol } $hit_of_scaff{ $key }[ $i ]{ 'hit' }, "\n";
					print { $fh_final } $hit_of_scaff{ $key }[ $i ]{ 'hit' }, "\n";
					$first_ol = 0;
				}
				print { $fh_ol } $hit_of_scaff{ $key }[ $j ]{ 'hit' }, "\n";
			}
		}
		# Hit was only counted if it had a duplicate -> count it once
		if ( $first ){
			$count_of{ $hit_of_scaff{ $key }[ $i ]{ 'id' } } += 1;
		}
		if ( $first and $first_ol ) {
			print { $fh_final } $hit_of_scaff{ $key }[ $i ]{ 'hit' }, "\n";
		}
	}
}

# Print count for each Rfam into fh_id after iterating through all scaffold
foreach my $key ( sort keys %count_of ) {
	print { $fh_id } $key, "\t", $count_of{ $key }, "\n";
}

close $fh_dupl;
close $fh_id;
close $fh_ol;

exit;
my $file_seq = "/home/tanja/holo_genomes/non_i5k/Amel_4.5_scaffolds.fa";

my %sequence_of;

open my $FH_seq, '<', $file_seq or die "Couldn't open file \"$file_seq\": $!\n";
my $header;

while ( my $line = <$FH_seq> ) {
	chomp $line;
	
	if ( $line =~ m/^>/ ) {
		$line =~ m/^>(.*)/;
		$header = $1;
	}
	else  {
		$sequence_of{ $header } = $line;
	}
}

close $FH_seq;

open my $fh_sorted, '<', $file_final or die "Couldn't open file \"$file_final\": $!\n";

while ( my $line = <$fh_sorted> ) {
	chomp $line;
	#		0		  1			2		   3	  4		5		 6		7		8		9	  10   11  12  13    14		15	  16	17
	# target_name accession query_name accession mdl mdl_from mdl_to seq_from seq_to strand trunc pass gc bias score e-value inc description
	my @parts = split( " ", $line );
	
	my $start;
	my $end;
	if ( $parts[ 9 ] =~ m/\+/ ) {
		$start = $parts[ 7 ];
		$end   = $parts[ 8 ];
	}
	else {
		$start = $parts[ 8 ];
		$end   = $parts[ 7 ];
	}
	my $dir = tempdir ( CLEANUP => 1 );
	my ( $fh, $filename ) = tempfile ( DIR => $dir, SUFFIX => '.fa' );
	my ( $fh_cm, $filename_cm ) = tempfile ( DIR => $dir, SUFFIX => '.cm' );
	
	my $length = $end - $start;
	my $seq_part = substr $sequence_of{ $parts[ 0 ] }, $start, $length;
	print { $fh } ">", $parts[ 5 ], "\n", $seq_part, "\n";
	my $dir_out = "/home/tanja/Rfam_11/amel_cmsearch_aln/";
	my $cm = $parts[ 3 ];
	# Familien alns in /home/tanja/Rfam_11/cms_sep_aln/
	my $cm_fetch = `cmfetch -o $filename_cm /home/tanja/Rfam_11/Rfam.cm.1_1 $cm`;
	
	my $aln = `cmalign -o $dir_out$cm.stk -g --noprob --cpu 6 --mapali /home/tanja/Rfam_11/cms_sep_aln/$cm.stk $filename_cm $filename`;
}

close $fh_sorted;