#!/usr/bin/perl

use strict;
use warnings;
use Number::Range;
use Data::Dumper;

# reference species
my $species = "amel";
#my $species = "test";
# file including closest genes of a CNE cluster on both sides in cis direction
# na means either no gene there or closest one is in trans direction
my $gene_file_one  	  = "/home/tanja/holo_genomes_results/CNE/cne_closest_two_genes_all_lnc_10_" . $species . ".tab";

# output files of CNEr, reference species is always first
my $CNE_file_one   	  = "/home/tanja/holo_genomes_results/CNE/Amel_Aros/CNE_first_second_70_100.bed";
my $CNE_file_two   	  = "/home/tanja/holo_genomes_results/CNE/Amel_Nvit/CNE_first_second_70_100.bed";
my $CNE_file_three 	  = "/home/tanja/holo_genomes_results/CNE/Amel_Oabi/CNE_first_second_70_100.bed";
#my $CNE_file_one   = "/home/tanja/holo_genomes_results/CNE/Amel_Aros/CNE_second_first_70_100.bed";
#my $CNE_file_two   = "/home/tanja/holo_genomes_results/CNE/Aros_Nvit/CNE_first_second_70_100.bed";
#my $CNE_file_three = "/home/tanja/holo_genomes_results/CNE/Aros_Oabi/CNE_first_second_70_100.bed";
#my $CNE_file_one   = "/home/tanja/holo_genomes_results/CNE/Amel_Nvit/CNE_second_first_70_100.bed";
#my $CNE_file_two   = "/home/tanja/holo_genomes_results/CNE/Aros_Nvit/CNE_second_first_70_100.bed";
#my $CNE_file_three = "/home/tanja/holo_genomes_results/CNE/Nvit_Oabi/CNE_first_second_70_100.bed";
#my $CNE_file_one   = "/home/tanja/holo_genomes_results/CNE/Amel_Oabi/CNE_second_first_70_100.bed";
#my $CNE_file_two   = "/home/tanja/holo_genomes_results/CNE/Aros_Oabi/CNE_second_first_70_100.bed";
#my $CNE_file_three = "/home/tanja/holo_genomes_results/CNE/Nvit_Oabi/CNE_second_first_70_100.bed";

open my $fh_gene, 	   '<', $gene_file_one or die "Couldn't open file \"$gene_file_one\": $!\n";
open my $fh_cne_one,   '<', $CNE_file_one or die "Couldn't open file \"$CNE_file_one\": $!\n";
open my $fh_cne_two,   '<', $CNE_file_two or die "Coudln't open file \"$CNE_file_two\": $!\n";
open my $fh_cne_three, '<', $CNE_file_three or die "Couldn't open file \"$CNE_file_three\": $!\n";

my $cne_file_in_nvit = "/home/tanja/holo_genomes_results/CNE/cne_in_cluster_unique_amel.txt";
my $cne_file_in_aros = "/home/tanja/holo_genomes_results/CNE/cne_in_cluster_unique_aros.txt";
my $cne_file_in_amel = "/home/tanja/holo_genomes_results/CNE/cne_in_cluster_unique_nvit.txt";
my $cne_file_in_oabi = "/home/tanja/holo_genomes_results/CNE/cne_in_cluster_unique_oabi.txt";

open my $fh_cne_amel,  '<', $cne_file_in_amel or die "Couldn't open file \"$cne_file_in_amel\": $!\n";
open my $fh_cne_aros,  '<', $cne_file_in_aros or die "Couldn't open file \"$cne_file_in_aros\": $!\n";
open my $fh_cne_nvit,  '<', $cne_file_in_nvit or die "Couldn't open file \"$cne_file_in_nvit\": $!\n";
open my $fh_cne_oabi,  '<', $cne_file_in_oabi or die "Couldn't open file \"$cne_file_in_oabi\": $!\n";

# create lists with all CNEs of one species
my @CNEs_amel;
my @CNEs_aros;
my @CNEs_nvit;
my @CNEs_oabi;

my %CNEs_amel;
my %CNEs_aros;
my %CNEs_nvit;
my %CNEs_oabi;

my $header;
my $key_cluster;
my %cne_in_cluster_of;
while ( my $line = <$fh_cne_nvit> ) {
	next if $line =~ m/^#/i;
	chomp $line;
	if ( $line =~ m/^>/i ) {
		$header = $line;
		my @parts = split ( "\t", $line );
		$key_cluster = $parts[ 0 ] . "\t" . $parts[ 1 ] . "\t" . $parts[ 2 ];
	}
	else {
		push @{ $CNEs_nvit{ $header } }, $line;
		push @{ $cne_in_cluster_of { $key_cluster } }, $line;
	}
}
close $fh_cne_nvit;

while ( my $line = <$fh_cne_aros> ) {
	next if $line =~ m/^#/i;
	chomp $line;
	if ( $line =~ m/^>/i ) {
		$header = $line;
	}
	else {
		push @{ $CNEs_aros{ $header } }, $line;
	}
}
close $fh_cne_aros;

while ( my $line = <$fh_cne_amel> ) {
	next if $line =~ m/^#/i;
	chomp $line;
	if ( $line =~ m/^>/i ) {
		$header = $line;
	}
	else {
		push @{ $CNEs_amel{ $header } }, $line;
	}
}
close $fh_cne_amel;

while ( my $line = <$fh_cne_oabi> ) {
	next if $line =~ m/^#/i;
	chomp $line;
	if ( $line =~ m/^>/i ) {
		$header = $line;
	}
	else {
		push @{ $CNEs_oabi{ $header } }, $line;
	}
}
close $fh_cne_oabi;

# create list with CNEs of reference species with all the hits in other species to this one
my @CNEs;

while ( my $line = <$fh_cne_one> ) {
	chomp $line;
	next if $line =~ m/^track/i;
	push @CNEs, $line;
}
close $fh_cne_one;

while ( my $line = <$fh_cne_two> ) {
	chomp $line;
	next if $line =~ m/^track/i;
	push @CNEs, $line;
}
close $fh_cne_two;

while ( my $line = <$fh_cne_three> ) {
	chomp $line;
	next if $line =~ m/^track/i;
	push @CNEs, $line;
}
close $fh_cne_three;

# get infos about the gene and the cluster it belongs to
my @gene_cluster_info;
my %scaffold_found;
while ( my $line = <$fh_gene> ) {
	next if $line =~ m/^#/i;
	chomp $line;
#	print $line, "\n";
#	exit;
	my @parts = split ( "\t", $line );
	push @gene_cluster_info, { 'scaffold' => $parts[ 0 ], 'start' => $parts[ 1 ], 'end' => $parts[ 2 ], 'distance_scaffold' => $parts[ 4 ], 'gene_upstream' => $parts[ 5 ], 'gene_dist_upstream' => $parts[ 6 ], 'gene_downstream' => $parts[ 7 ], 'gene_dist_downstream' => $parts[ 8 ] };
	$scaffold_found{ $parts[ 0 ] } += 1;
}
close $fh_gene;

# updated CNE coordinates to final range of CNEs after comparing overlapping hits
my %CNEs_updated;
#my $count = 0;

# an Hash anpassen anstatt an Array, dann klappt auch die Zuordnung der Array-ID für den Hit
foreach my $CNE ( sort @CNEs ) {
	my @parts1 = split ( "\t", $CNE );
	my $scaffold = ">" . $parts1[ 0 ];
	next if !exists $scaffold_found{ $scaffold };
	print "CNE: ", $CNE, "\n";
	my $key_ref;
	foreach my $key ( sort keys %CNEs_nvit ) {
		my @parts2 = split ( "\t", $key );
		my $key_cluster = $parts2[ 0 ] . "\t" . $parts2[ 1 ] . "\t" . $parts2[ 2 ];
		if ( $parts2[ 0 ] eq ">" . $parts1[ 0 ] ) {
#			print "parts2: ", $parts2[ 1 ], "\t", $parts2[ 2 ], "\t", "parts1: ", $parts1[ 1 ], "\t", $parts1[ 2 ], "\n";
			foreach my $cluster ( @gene_cluster_info ) {
				if ( $parts2[ 0 ] eq $cluster->{ 'scaffold' } ) {
				#	my $range3 = Number::Range->new( $cluster->{ 'start' }..$cluster->{ 'end' } );
				#	my @inrange3 = $range3->inrange( $parts1[ 1 ], $parts1[ 2 ] );
				#	if ( $inrange3[ 0 ] && $inrange3[ 1 ] ) {	
				#	print $parts2[ 1 ], "\t", $parts2[ 2 ], "\n";
				#	print $parts1[ 1 ], "\t", $parts1[ 2 ], "\n";
				#	exit;
						my $range = Number::Range->new( $parts2[ 1 ]..$parts2[ 2 ] );
						my @inrange = $range->inrange( $parts1[ 1 ], $parts1[ 2 ] );
						if ( $inrange[ 0 ] or $inrange[ 1 ] ) {
							$key_ref = $parts2[ 0 ] . "\t" . $parts2[ 1 ] . "\t" . $parts2[ 2 ];
							foreach my $cne_unique ( @{ $cne_in_cluster_of{ $key_cluster } } ) {
						#		print "cne_unique: ", $cne_unique, "\n";
						#		exit;
								my @parts3 = split ( "\t", $cne_unique );
							#	my $range2 = Number::Range->new( $parts3[ 0 ]..$parts3[ 1 ] );
							#	my @inrange2 = $range2->inrange( $parts1[ 1 ], $parts1[ 2 ] );
								my $hit = $parts3[ 0 ] . "_" . $parts3[ 1 ];
						#		print "hit: ", $hit, "\n";
						#		exit;
								$CNEs_updated{ $key_ref }{ $hit }[ 0 ] = 'def';	
							}
						}
				#	}
				#	last;
				}
			}
		}
	}
#	print Dumper ( %CNEs_updated ), "\n";
#			exit;
	next if !$key_ref;
	my @parts3 = split ( ":", $parts1[ 3 ] );
	my @parts4 = split ( "-", $parts3[ 1 ] );

	if ( $CNE =~ m/aros/i ) {
#	if ( $CNE =~ m/drei/i ) {
		foreach my $key ( sort keys %CNEs_aros ) {
			my @parts2 = split ( "\t", $key );
			if ( $parts2[ 0 ] eq ">" . $parts3[ 0 ] ) {
				foreach my $cne_aros ( @{ $CNEs_aros{ $key } } ) {
					my @parts5 = split ( "\t", $cne_aros );
#					print "aros: ", $cne_aros, "\n";
#					exit;
					my $range = Number::Range->new( $parts5[ 0 ]..$parts5[ 1 ] );
					my @inrange = $range->inrange( $parts4[ 0 ], $parts4[ 1 ] );
					if ( $inrange[ 0 ] or $inrange[ 1 ] ) {
						my $hit = $parts2[ 0 ] . "_" . $parts5[ 0 ] . "_" . $parts5[ 1 ];
						foreach my $key2 ( sort keys $CNEs_updated{ $key_ref } ) {
							my @coords = split "_", $key2;
							my $range2 = Number::Range->new( $coords[ 0 ]..$coords[ 1 ] );
							my @inrange2 = $range2->inrange( $parts1[ 1 ], $parts1[ 2 ] );
					#		print "@inrange2", "\n";
					#		print $hit, "\n";
					#		print $key2, "\n";
							#exit;
							push  ( @{ $CNEs_updated{ $key_ref }{ $key2 } }, $hit ) if $inrange2[ 0 ] or $inrange2[ 1 ];
#							last if $inrange2[ 0 ] or $inrange2[ 1 ];
					#		print Dumper ( %CNEs_updated ), "\n";
#							exit;
						}
					}
				}
			}
		 }
	}
	elsif ( $CNE =~ m/nvit/i ) {
#	elsif ( $CNE =~ m/zwei/i ) {
		foreach my $key ( sort keys %CNEs_amel ) {
			my @parts2 = split ( "\t", $key );
			if ( $parts2[ 0 ] eq ">" . $parts3[ 0 ] ) {
				foreach my $cne_amel ( @{ $CNEs_amel{ $key } } ) {
#					print "amel: ", $cne_amel, "\n";
#					exit;
					my @parts5 = split ( "\t", $cne_amel );
					my $range = Number::Range->new( $parts5[ 0 ]..$parts5[ 1 ] );
					my @inrange = $range->inrange( $parts4[ 0 ], $parts4[ 1 ] );
					if ( $inrange[ 0 ] or $inrange[ 1 ] ) {
						my $hit = $parts2[ 0 ] . "_" . $parts5[ 0 ] . "_" . $parts5[ 1 ];
						foreach my $key2 ( sort keys $CNEs_updated{ $key_ref } ) {
							my @coords = split "_", $key2;
							my $range2 = Number::Range->new( $coords[ 0 ]..$coords[ 1 ] );
							my @inrange2 = $range2->inrange( $parts1[ 1 ], $parts1[ 2 ] );
#							print "amel: ", $cne_amel, "\n";
#							print $key, "\n";
#							print "key2: ", $key2, "\n" if $inrange2[ 0 ] or $inrange2[ 1 ];
#							exit if $inrange2[ 0 ] or $inrange2[ 1 ];
							push  ( @{ $CNEs_updated{ $key_ref }{ $key2 } }, $hit ) if $inrange2[ 0 ] or $inrange2[ 1 ];
#							last if $inrange2[ 0 ] or $inrange2[ 1 ];
						}
					}
				}
			}
		}
	}
	elsif ( $CNE =~ m/oabi/i ) {
#	elsif ( $CNE =~ m/vier/i ) {
		foreach my $key ( sort keys %CNEs_oabi ) {
			my @parts2 = split ( "\t", $key );
			if ( $parts2[ 0 ] eq ">" . $parts3[ 0 ] ) {
				foreach my $cne_oabi ( @{ $CNEs_oabi{ $key } } ) {
#					print "key: ", $key, "\n";
#					print "oabi: ", $cne_oabi, "\n";
#					exit;
					my @parts5 = split ( "\t", $cne_oabi );
					my $range = Number::Range->new( $parts5[ 0 ]..$parts5[ 1 ] );
					my @inrange = $range->inrange( $parts4[ 0 ], $parts4[ 1 ] );
					if ( $inrange[ 0 ] or $inrange[ 1 ] ) {
						my $hit = $parts2[ 0 ] . "_" . $parts5[ 0 ] . "_" . $parts5[ 1 ];
						foreach my $key2 ( sort keys $CNEs_updated{ $key_ref } ) {
							my @coords = split ( "_", $key2);
							my $range2 = Number::Range->new( $coords[ 0 ]..$coords[ 1 ] );
							my @inrange2 = $range2->inrange( $parts1[ 1 ], $parts1[ 2 ] );
							push  ( @{ $CNEs_updated{ $key_ref }{ $key2 } }, $hit ) if $inrange2[ 0 ] or $inrange2[ 1 ];
#							last if $inrange2[ 0 ] or $inrange2[ 1 ];
						}
					}
				}
			}
		}
	}
#	last;
 }

print Dumper( %CNEs_updated ), "\n";
#exit;

my @gene_cne_cluster_info;

foreach my $cluster ( @gene_cluster_info ) {
	print $cluster->{ 'scaffold' }, "\n";
	my $key = $cluster->{ 'scaffold' } . "\t" . $cluster->{ 'start' } . "\t" . $cluster->{ 'end' };
	my @cne_hits;
#	print "key: ", $key, "\n";
	foreach my $key_ref ( sort keys %CNEs_updated ) {
#		print "key_ref: ", $key_ref, "\n";
		if ( $key_ref eq $key ) {
			my @parts = split ( "\t", $key_ref );
			foreach my $key2 ( sort keys $CNEs_updated{ $key_ref } ) {
#				print "key2: ", $key2, "\n";
				my @coords_cluster = split ( "_", $key2 );
				#my $first = shift @{ $CNEs_updated{ $key_ref }{ $key2 } };
				my @sorted_items = sort @{ $CNEs_updated{ $key_ref }{ $key2 } };
				my $range = Number::Range->new( $parts[ 1 ]..$parts[ 2 ] );
				my @inrange = $range->inrange( $coords_cluster[ 0 ], $coords_cluster[ 1 ] );
				if ( $inrange[ 0 ] or $inrange[ 1 ] ) {
					for ( my $i = 0; $i <  @{ $cne_in_cluster_of{ $key_ref } }; ++$i ) {
						print "cne_in_cluster_key_ref_i: ", $cne_in_cluster_of{ $key_ref }[ $i ], "\n";
						my @cne_coords = split ( "\t", $cne_in_cluster_of{ $key_ref }[ $i ] );
						if ( $cne_coords[ 0 ] == $coords_cluster[ 0 ] && $cne_coords[ 1 ] == $coords_cluster[ 1 ] ) {
							print "match\n";
							for ( my $j = 0; $j < @sorted_items; ++$j ) {
								next if $sorted_items[ $j ] eq 'def';
								my $hit = $sorted_items[ $j ] . "_$i";
								print $hit, "\n";
								push @cne_hits, $hit;
							#	exit;
							}
#							print Dumper ( @cne_hits ), "\n";
#							exit;
						}
					}
				#	exit;
				}
			}
		}
	}
#exit;
	#hier die hits irgendwie an reihenfolge anpassen. geht das an der stelle schon? oder später wenn ich für alle arten die reihenfolgeninfo habe?
	# mit [#] die wobei # sich auf die positions des cnes im cluster bezieht?
	push @gene_cne_cluster_info, { 'scaffold' => $cluster->{ 'scaffold' }, 'start' => $cluster->{ 'start' }, 'end' => $cluster->{ 'end' }, 'distance_scaffold' => $cluster->{ 'distance_scaffold' }, 'gene_upstream' => $cluster->{ 'gene_upstream' }, 'gene_dist_upstream' => $cluster->{ 'gene_dist_upstream' }, 'gene_downstream' => $cluster->{ 'gene_downstream' }, 'gene_dist_downstream' => $cluster->{ 'gene_dist_downstream' }, 'cne_hits' => \@cne_hits };
}

#exit;
my $file_out = "/home/tanja/holo_genomes_results/CNE/cne_closest_two_genes_all_hitinfo_lnc_10_nodups_" . $species . "2.tab";
open my $fh_out, '>', $file_out or die "Couldn't open file \"$file_out\": $!\n";

@gene_cne_cluster_info = sort @gene_cne_cluster_info;

# print Dumper ( @gene_cne_cluster_info ), "\n";

print { $fh_out } "#Scaffold\tCluster_start\tCluster_end\tClosest_gene_upstream\tDistance_to_gene_upstream\tClosest_gene_downstream\tDistance_to_gene_downstream\tCNE_hits_in_other_species\n";

foreach my $cluster ( @gene_cne_cluster_info ) {
	print { $fh_out } $cluster->{ 'scaffold' }, "\t", $cluster->{ 'start' }, "\t", $cluster->{ 'end' }, "\t", $cluster->{ 'gene_upstream' }, "\t", $cluster->{ 'gene_dist_upstream' }, "\t", $cluster->{ 'gene_downstream' }, "\t", $cluster->{ 'gene_dist_downstream' }, "\t";
#	print Dumper ( $cluster->{ 'cne_hits' } ), "\n";
#	exit;
	for ( my $i = 0; $i < @{ $cluster->{ 'cne_hits' } }; ++$i ) {
		print { $fh_out } $cluster->{ 'cne_hits' }[ $i ] , ", ";
	}
	print { $fh_out } "\n";
}

close $fh_out;

exit;