#!/usr/bin/perl

use strict;
use warnings;
use Number::Range;

# species abbreviation, same as in cne_gene_neighbourhood.pl
my $species = 'aros';

# input file with all CNEs in a cluster listed
my $file_in = "/home/tanja/holo_genomes_results/CNE/cne_in_cluster_". $species . ".txt";
open my $fh_in, '<', $file_in or die "Couldn't open file \"$file_in\": $!\n";

my %cne_cluster;
my $cluster_id;

# create hash with cluster info as key, all CNEs pushed as value in a list
while ( my $line = <$fh_in> ) {
	next if $line =~ m/^Scaffold/;
	chomp $line;
	if ( $line =~ m/^>/ ){
		$cluster_id = $line;
	}
	else {
		push @{ $cne_cluster{ $cluster_id } }, $line;
	}
}
close $fh_in;

# go through results of cne_gene_neighbourhood and combine overlooked overlapping cluster
my %cne_cluster_unique;
my $change = 0;
foreach my $key ( sort keys %cne_cluster ) {
	my @parts1 = split ( "\t", $key );
	if ( ! %cne_cluster_unique ) {
		for ( my $i = 0; $i < @{ $cne_cluster{ $key } }; ++$i ) {
			push @{ $cne_cluster_unique{ $key } }, $cne_cluster{ $key }[ $i ];
		}
	}
	else {
		my $found = 0;
		foreach my $key2 ( sort keys %cne_cluster_unique ) {
			next if $key eq $key2;
			my @parts2 = split ( "\t", $key2 );
			if ( $parts1[ 0 ] eq $parts2[ 0 ] ) {
				my $range1 = Number::Range->new( $parts1[ 1 ]..$parts1[ 2 ] );
				my $range2 = Number::Range->new( $parts2[ 1 ]..$parts2[ 2 ] );
				my @inrange1 = $range1->inrange( $parts2[ 1 ], $parts2[ 2 ] );
				my @inrange2 = $range2->inrange( $parts1[ 1 ], $parts1[ 2 ] );
				my $distance;
				my $start = $parts1[ 1 ];
				my $end   = $parts1[ 2 ];
				if ( $parts1[ 2 ] < $parts2[ 1 ] ) {
					$distance = abs( $parts2[ 1 ] - $parts1[ 2 ] );
				}
				else {
					$distance = abs( $parts1[ 1 ] - $parts2[ 2 ] );
				}
				if ( $inrange1[ 0 ] or $inrange1[ 1 ] or $inrange2[ 0 ] or $inrange2[ 1 ] or $distance <= 20000) {	
					for ( my $i = 0; $i < @{ $cne_cluster{ $key } }; ++$i ) {
#						print "cne_cluster: ", $cne_cluster{ $key }[ $i ], "\n";
#						print "cne cluster unique: ", $cne_cluster_unique{ $key2 }[ -1 ], "\n";
						push @{ $cne_cluster_unique{ $key2 } }, $cne_cluster{ $key }[ $i ];
					}
					$start = $parts2[ 1 ] if $parts2[ 1 ] < $start;
					$end   = $parts2[ 2 ] if $parts2[ 2 ] > $end;
					my $key3 = $parts1[ 0 ] . "\t". $start . "\t" . $end . "\t" . scalar @{ $cne_cluster_unique{ $key2 } };
					$cne_cluster_unique{ $key3 } = delete $cne_cluster_unique{ $key2 };
					$found  = 1;
					$change = 1;
				}
			}
			
		}
		if ( !$found ) {
			for ( my $i = 0; $i < @{ $cne_cluster{ $key } }; ++$i ) {
				push @{ $cne_cluster_unique{ $key } }, $cne_cluster{ $key }[ $i ];
			}
		}
	}
}

my %unique_cne_cluster;

# remove redundant CNEs but adapt start/stop coordinates
# adapt start/stop/cne count in key
foreach my $key ( sort keys %cne_cluster_unique ) {
	if ( scalar @{ $cne_cluster_unique{ $key } } == 1 ) {
		push @{ $unique_cne_cluster{ $key } }, $cne_cluster_unique{ $key }[ 0 ];
	}
	else {
		@{ $cne_cluster_unique{ $key } } = sort @{ $cne_cluster_unique{ $key } };
		# neuen key mit richtigen Koordinaten erstellen, value beim alten löschen und dann in späterem Schritt alle keys ohne value löschen?
		for ( my $i = 0; $i < @{ $cne_cluster_unique{ $key } }; ++$i ) {
			my @parts_i = split ( "\t", $cne_cluster_unique{ $key }[ $i ] );
			if ( exists $unique_cne_cluster{ $key } ) {
				my $found = 0;
#				print $key, "\t", $unique_cne_cluster{ $key }, "\n";
				for ( my $j = 0; $j < @{ $unique_cne_cluster{ $key } }; ++$j ) {
#					print $unique_cne_cluster{ $key }[ $j ], "\n";
					exit if !$unique_cne_cluster{ $key }[ $j ];
					my @parts_j = split( "\t", $unique_cne_cluster{ $key }[ $j ] );
					my $range1 = Number::Range->new( $parts_i[ 0 ]..$parts_i[ 1 ] );
					my $range2 = Number::Range->new( $parts_j[ 0 ]..$parts_j[ 1 ] );
					my @inrange1 = $range1->inrange( $parts_j[ 0 ], $parts_j[ 1 ] );
					my @inrange2 = $range2->inrange( $parts_i[ 0 ], $parts_i[ 1 ] );
					if ( $inrange1[ 1 ] == 1 && $inrange2[ 1 ] == 0  ) {
						$unique_cne_cluster{ $key }[ $j ] = $parts_j[ 0 ] . "\t". $parts_i[ 1 ];
						$found = 1;
						last;
					}
					elsif ( $inrange2[ 0 ] == 1 && $inrange2[ 1 ] == 1 ) {
						$found = 1;
						last;
					}
				}
				if ( !$found ) {
					push @{ $unique_cne_cluster{ $key } }, $cne_cluster_unique{ $key }[ $i ];
				}
			}
			else {
				push @{ $unique_cne_cluster{ $key } }, $cne_cluster_unique{ $key }[ $i ];
			}
		}
	}
}
print "cne_cluster: ", scalar keys %cne_cluster, "\n";
# create file with all unique CNEs, same structure as input file
my $file_out = "/home/tanja/holo_genomes_results/CNE/cne_in_cluster_unique_" . $species . ".txt";
open my $fh_out, '>', $file_out or die "Couldn't open file \"$file_out\": $!\n";

print { $fh_out } "#Scaffold\tCluster_start\tCluster_end\tNo_of_CNEs_in_cluster\n";
# remove redundant left overs from unique hash
foreach my $key ( sort keys %unique_cne_cluster ) {
	my @parts = split ( "\t", $key );
	my $start = $parts[ 1 ];
	my $end   = $parts[ 2 ];
	for ( my $i = 0; $i < @{ $unique_cne_cluster{ $key } }; ++$i ) {
		my @parts2 = split( "\t", $unique_cne_cluster{ $key }[ $i ] );
		$start = $parts2[ 0 ] if $parts2[ 0 ] < $start;
		$end   = $parts2[ 1 ] if $parts2[ 1 ] > $end;
	}
	my $key2 = $parts[ 0 ] . "\t". $start . "\t" . $end . "\t" . scalar @{ $unique_cne_cluster{ $key } };
	$unique_cne_cluster{ $key2 } = delete $unique_cne_cluster{ $key };
}

# fasta file of reference species
my $genome_file = "/home/tanja/holo_genomes/Aros01112013-genome.fa";
#my $genome_file = "/home/tanja/holo_genomes/Oabi11242013.genome.fa";
#my $genome_file = "/home/tanja/holo_genomes/GCA_000002325.2_Nvit_2.1_genomic.fna";
#my $genome_file = "/home/tanja/holo_genomes/GCA_000002195.1_Amel_4.5_genomic.fna";

my $scaffold_length = &getScaffoldlength( $genome_file, $species );

my $annotation_file_all = "/home/tanja/holo_genomes/ATHROS_OGS_with_lnc_sorted.gff";
#my $annotation_file_all = "/home/tanja/holo_genomes/ORUABI_OGSv1-0_withlnc_sorted.gff";
#my $annotation_file_all = "/home/tanja/holo_genomes/ref_Nvit_2.1_top_level_genebank.gff3";
#my $annotation_file_all = "/home/tanja/holo_genomes/Amel_4.5_top_level_mod_cner.gff3";

my $gene_info_of = &readAnnotation( $annotation_file_all, $species );

my $closest_gene_output = "/home/tanja/holo_genomes_results/CNE/cne_closest_gene_unique_" . $species. ".tab";
open my $fh_closest_gene, '>', $closest_gene_output or die "Couldn't open file \"$closest_gene_output\": $!\n";

print { $fh_closest_gene } "# Scaffold\tCluster_start\tCluster_stop\tCNE_number_in_cluster\tDistance_to_Scaffold_end\n";

# print unique CNEs to file
foreach my $key ( sort keys %unique_cne_cluster ) {
	print $key, "\n";
	my @parts = split( "\t", $key );
	print { $fh_out } $key, "\n";
	for ( my $i = 0; $i < @{ $unique_cne_cluster{ $key } }; ++$i ) {
		print { $fh_out } $unique_cne_cluster{ $key }[ $i ], "\n";
	}
	my @genes_upstream;
	my @genes_downstream;
	my @genes_in_cluster;
	# add corresponding genes of cluster to hash
	foreach my $key3 ( sort keys $gene_info_of ) {
		next if $key =~ m/^TRNA/i;
		next if $key =~ m/^Mir/i;
		if ( $parts[ 0 ] eq $gene_info_of->{ $key3 }[ 0 ]{ 'scaffold' } ) {
			if ( $gene_info_of->{ $key3 }[ 0 ]{ 'end' } < $parts[ 1 ] && ( $parts[ 1 ] - $gene_info_of->{ $key3 }[ 0 ]{ 'end' } ) < 500000 ) {
				push @genes_upstream, $key3;
			}
			elsif ( ( $gene_info_of->{ $key3 }[ 0 ]{ 'start' } >= $parts[ 1 ] && $gene_info_of->{ $key3 }[ 0 ]{ 'start' } < $parts[ 2 ] ) or ( $gene_info_of->{ $key3 }[ 0 ]{ 'end' } > $parts[ 1 ] && $gene_info_of->{ $key3 }[ 0 ]{ 'end' } <= $parts[ 2 ] ) ) {
				push @genes_in_cluster, $key3;
			}
			elsif ( $gene_info_of->{ $key3 }[ 0 ]{ 'start' } > $parts[ 2 ] && ( $gene_info_of->{ $key3 }[ 0 ]{ 'start' } - $parts[2 ] ) < 500000 ) {
				push @genes_downstream, $key3;
			}
		}
	}
	# add na if no gnes could be found next to/in a cluster
	if ( !@genes_upstream ) {
		push @genes_upstream, 'na';
	}
	if ( !@genes_downstream ) {
		push @genes_downstream, 'na';
	}
	if ( !@genes_in_cluster ) {
		push @genes_in_cluster, 'na';
	}
	print { $fh_closest_gene } $key, "\t";
	my $distance = $scaffold_length->{ $parts[ 0 ] } - $parts[ 2 ];
	print { $fh_closest_gene } $distance;
	print { $fh_closest_gene } "\nGenes upstream: \n";
	foreach my $gene ( @genes_upstream ) {
		print { $fh_closest_gene } $gene, "\t";
	}
	print { $fh_closest_gene } "\nGenes downstream: \n";
	foreach my $gene ( @genes_downstream ) {
		print { $fh_closest_gene } $gene, "\t";
	}
	print { $fh_closest_gene } "\nGenes in cluster: \n";
	foreach my $gene ( @genes_in_cluster ) {
		print { $fh_closest_gene } $gene, "\t";
	}
	print { $fh_closest_gene } "\n";
}
print "unique_cluster_new: ", scalar keys %unique_cne_cluster, "\n";
close $fh_out;
close $fh_closest_gene;

exit;

### Subroutine(s) ###

# adds species abbreviation to scaffolds
sub addSpecies {
	my $sequence_of = shift @_;
	my $species 	= shift @_;

	my %species_sequence_of;
	
	foreach my $key (sort keys $sequence_of ) {
		$key =~ m/>(\S*).*/;
		my $key2 = ">" . $species . "_" . $1;
		$species_sequence_of{ $key2 } = $sequence_of->{ $key };
	}
	return \%species_sequence_of;
}

sub readAnnotation {
	my $annotation_file = shift @_;
	my $species 		= shift @_;
	open my $fh_annotation, '<', $annotation_file or die "Couldn't open file: \"$annotation_file\": $!\n";
	my %gene_info_of;
	while ( my $line = <$fh_annotation> ) {
		chomp $line;
		if ( $line =~ m/\tgene\t/ or $line =~ m/\ttranscript\t/ ) {
			#Spalte 1 = Scaffold, Spalte 4 = start, Spalte 5 = stop, Spalte 7 = Strang, Spalte 9 = Genname + anderes (Name oder GeneID als Identifier?)
			my @parts    = split( "\t", $line );
			$parts[ 0 ] =~ m/[a-z]*_?(.*)/i;
			my $scaffold = ">" . $species . "_" . $parts[ 0 ];
			#my $scaffold = ">" . $species . "_" . $1;
			my $start    = $parts[ 3 ];
			my $end      = $parts[ 4 ];
			my $strand   = $parts[ 6 ];
			# Nasonia und Apis
			#$parts[ 8 ]  =~ m/name=([^;]+);/i;
			# Orussus und Athalia
			$parts[ 8 ] =~ m/ID=([^;]+);/i;
			my $genename = $1;
			push @{ $gene_info_of{ $genename } }, { 'scaffold' => $scaffold, 'start' => $start, 'end' => $end, 'strand' => $strand };
		}
	}
	close $fh_annotation;
	return \%gene_info_of;
}

sub getScaffoldlength {
	my $genome_file = shift @_;
	my $species 	= shift @_;
	my $sequence_of = &readFasta( $genome_file );
	$sequence_of 	= &addSpecies( $sequence_of, $species );

	my %scaffold_length;
	foreach my $key ( keys $sequence_of ){
		my $length = length $sequence_of->{ $key };
		$scaffold_length{ $key } = $length;
	}
	return \%scaffold_length;
}

# read fasta file into hash, header as key
sub readFasta {
	my $genome_file = shift @_;
	open my $fh_genome, '<', $genome_file or die "Couldn't open file: \"$genome_file\": $!\n";

	my %sequence_of;
	my $first_header = 1;
	my $header;
	my $sequence;

	while ( my $line = <$fh_genome> ) {
		chomp $line;
		if ( $line =~ m/^>/ ) {
			if ( $first_header ) {
				$header = $line;
				$first_header = 0;
				next;
			}
			$sequence_of{ $header } = $sequence;
			$header = $line;
			$sequence = '';
		}
		else {
			$sequence .= $line;
		}
	}
	close $fh_genome;
	return \%sequence_of;
}