#!/usr/bin/perl

use strict;
use warnings;
use Number::Range;

# annotation file of reference species
# make sure the scaffold names are the same in gff, fasta and CNEr output (problem with gene bank, ref seq and other databases, e.g. beebase)
# für Nvit readAnnotation anpassen, weil Artkürzel schon im gff mit drin
#my $annotation_file_all = "/home/tanja/holo_genomes/ATHROS_OGS_with_lnc_sorted.gff";
my $annotation_file_all = "/home/tanja/holo_genomes/ORUABI_OGSv1-0_withlnc_sorted.gff";
#my $annotation_file_all = "/home/tanja/holo_genomes/ref_Nvit_2.1_top_level_genebank.gff3";
#my $annotation_file_all = "/home/tanja/holo_genomes/Amel_4.5_top_level_mod_cner.gff3";

# species abbreviation of reference species
#my $species = "aros";
my $species = "oabi";
#my $species = "nvit";
#my $species = "amel";

# get all gene + lncRNA names of the species
my $gene_info_of = &readAnnotation( $annotation_file_all, $species );

# CNE files
# output files of CNEr, reference species always first in file
#my $CNEs_both_first_file  = "/home/tanja/holo_genomes_results/CNE/Aros_Oabi/CNE_first_second_70_100.bed";
#my $CNEs_both_second_file = "/home/tanja/holo_genomes_results/CNE/Aros_Nvit/CNE_first_second_70_100.bed";
#my $CNEs_both_third_file  = "/home/tanja/holo_genomes_results/CNE/Amel_Aros/CNE_second_first_70_100.bed";

my $CNEs_both_first_file  = "/home/tanja/holo_genomes_results/CNE/Aros_Oabi/CNE_second_first_70_100.bed";
my $CNEs_both_second_file = "/home/tanja/holo_genomes_results/CNE/Nvit_Oabi/CNE_second_first_70_100.bed";
my $CNEs_both_third_file  = "/home/tanja/holo_genomes_results/CNE/Amel_Oabi/CNE_second_first_70_100.bed";

#my $CNEs_both_first_file  = "/home/tanja/holo_genomes_results/CNE/Nvit_Oabi/CNE_first_second_70_100_test.bed";
#my $CNEs_both_second_file = "/home/tanja/holo_genomes_results/CNE/Aros_Nvit/CNE_second_first_70_100_test.bed";
#my $CNEs_both_third_file  = "/home/tanja/holo_genomes_results/CNE/Amel_Nvit/CNE_second_first_70_100_test.bed";

#my $CNEs_both_first_file  = "/home/tanja/holo_genomes_results/CNE/Amel_Oabi/CNE_first_second_70_100.bed";
#my $CNEs_both_second_file = "/home/tanja/holo_genomes_results/CNE/Amel_Nvit/CNE_first_second_70_100.bed";
#my $CNEs_both_third_file  = "/home/tanja/holo_genomes_results/CNE/Amel_Aros/CNE_first_second_70_100.bed";

# read CNE file
# columns 1-3 without Header important [Scaffold Start Stop]
my $CNEs_both_first  = &readCNEs( $CNEs_both_first_file );
my $CNEs_both_second = &readCNEs( $CNEs_both_second_file );
my $CNEs_both_third  = &readCNEs( $CNEs_both_third_file );

# fasta file of reference species
#my $genome_file = "/home/tanja/holo_genomes/Aros01112013-genome.fa";
my $genome_file = "/home/tanja/holo_genomes/Oabi11242013.genome.fa";
#my $genome_file = "/home/tanja/holo_genomes/GCA_000002325.2_Nvit_2.1_genomic.fna";
#my $genome_file = "/home/tanja/holo_genomes/GCA_000002195.1_Amel_4.5_genomic.fna";

#output file for CNE sequences
my $cne_sequence_output = "/home/tanja/holo_genomes_results/CNE/cne_sequence_". $species . ".fa";

# get nucleotide sequence of each CNE of reference species from fasta file
&getCNEsequence( $CNEs_both_first,  $genome_file, $cne_sequence_output, $species );
&getCNEsequence( $CNEs_both_second, $genome_file, $cne_sequence_output, $species );
&getCNEsequence( $CNEs_both_third,  $genome_file, $cne_sequence_output, $species );

# get length of each scaffold, used in distance calculation to scaffold end
my $scaffold_length = &getScaffoldlength ( $genome_file, $species );

# identify CNE cluster, distance < 20,000 between CNEs = same cluster
# CNEs innerhalb des Clusters müssen noch aufsteigend sortiert werden!
my $CNE_cluster = &getCNEcluster( $CNEs_both_first, $CNEs_both_second, $CNEs_both_third );

foreach my $key (sort keys $CNE_cluster ) {
	print $key, "\n";
	my $start;
	my $end;
	my @delete_list;
	foreach my $key2 ( sort keys $CNE_cluster->{ $key } ) {
		print "key2: ", $key2, "\n";
		$start = $CNE_cluster->{ $key }->{ $key2 }[ 0 ]{ 'start' };
		$end   = $CNE_cluster->{ $key }->{ $key2 }[ 0 ]{ 'end' };
		print $start, "\t", $end, "\n";
		my $count = keys $CNE_cluster->{ $key };
		if ( $count > 1 ) {
			for ( my $i = 1; $i < @{ $CNE_cluster->{ $key }->{ $key2 } }; ++$i ) {
				$start = $CNE_cluster->{ $key }->{ $key2 }[ $i ]{ 'start' } if $CNE_cluster->{ $key }->{ $key2 }[ $i ]{ 'start' } < $start;
				$end   = $CNE_cluster->{ $key }->{ $key2 }[ $i ]{ 'end' }   if $CNE_cluster->{ $key }->{ $key2 }[ $i ]{ 'end' } > $end;
			}
		}
		print $start, "\t", $end, "\n";
		foreach my $key3 (sort keys $CNE_cluster->{ $key } ) {
			next if $key3 eq $key2;
			my $start2 = $CNE_cluster->{ $key }->{ $key3 }[ 0 ]{ 'start' };
			my $end2   = $CNE_cluster->{ $key }->{ $key3 }[ 0 ]{ 'start' };
			for ( my $i = 1; $i < @{ $CNE_cluster->{ $key }->{ $key3 } }; ++$i ) {
				$start2 = $CNE_cluster->{ $key }->{ $key3 }[ $i ]{ 'start' } if $CNE_cluster->{ $key }->{ $key3 }[ $i ]{ 'start' } < $start2;
				$end2   = $CNE_cluster->{ $key }->{ $key3 }[ $i ]{ 'end' }   if $CNE_cluster->{ $key }->{ $key3 }[ $i ]{ 'end' } > $end2;
			}
			my $range1   = Number::Range->new( $start..$end );
			my $range2   = Number::Range->new( $start2..$end2 );
			my @inrange1 = $range1->inrange( $start2, $end2 );
			my @inrange2 = $range2->inrange( $start, $end );
			print $start2, "\t", $end2, "\n";
			if ( $inrange1[ 0 ] or $inrange1[ 1 ] or $inrange2[ 0 ] or $inrange2[ 1 ] ) {
				push @delete_list, $key3;
				for ( my $i = 0; $i < @{ $CNE_cluster->{ $key }->{ $key3 } }; ++$i ) {
					print $CNE_cluster->{ $key }->{ $key3 }[ $i ]{ 'start' }, "\n";
					print $CNE_cluster->{ $key }->{ $key2 }[ 0 ]{ 'start' }, "\n";
					push @{ $CNE_cluster->{ $key }->{ $key2 } }, { 'start' => $CNE_cluster->{ $key }->{ $key3 }[ $i ]{ 'start' }, 'end' => $CNE_cluster->{ $key }->{ $key3 }[ $i ]{ 'end' } };
				}
			}
		}
	}
#	print "Anzahl keys vorher: ", scalar keys $CNE_cluster->{ $key }, "\n";
#	foreach my $key3 ( @delete_list ) {
#		print $key3, "\n";
#		print $CNE_cluster->{ $key }->{ $key3 }[ 0 ]{ 'start' }, "\n";
#		print $CNE_cluster->{ $key }->{ $key3 }, "\n";
#		delete $CNE_cluster->{ $key }->{ $key3 };
#		print $CNE_cluster->{ $key }->{ $key3}, "\n";
#	}
#	print "Anzahl keys nachher: ", scalar keys $CNE_cluster->{ $key }, "\n";
#	last;
}
# get genes on both sides of CNE cluster with distance < 500 kb to end of scaffold
my $closest_genes_of = &getclosestgenes( $CNE_cluster, $gene_info_of, $species );

# write CNE infos (scaffold, start, stop, CNE no in cluster, distance to scaffold end) to file
# followed by closest genes upstream, downstream and in cluster
my $closest_gene_output = "/home/tanja/holo_genomes_results/CNE/cne_closest_gene_" . $species. ".tab";
open my $fh_closest_gene, '>', $closest_gene_output or die "Couldn't open file \"$closest_gene_output\": $!\n";

print { $fh_closest_gene } "# Scaffold\tCluster_start\tCluster_stop\tCNE_number_in_cluster\tDistance_to_Scaffold_end\n";

foreach my $key ( sort keys $closest_genes_of ) {
	foreach my $key2 ( sort keys $closest_genes_of->{ $key } ) {
		print { $fh_closest_gene } ">$key\t";
		foreach my $key3 ( sort { $b cmp $a } keys $closest_genes_of->{ $key }->{ $key2 }[ 0 ] ) {
			print { $fh_closest_gene } $closest_genes_of->{ $key }->{ $key2 }[ 0 ]{ $key3 }, "\t";
		}
		my $distance = $scaffold_length->{ $key } - $closest_genes_of->{ $key }->{ $key2 }[ 0 ]{ 'end' };
		print { $fh_closest_gene } $distance;
		print { $fh_closest_gene } "\nGenes upstream: \n";
		for ( my $i = 0; $i < @{ $closest_genes_of->{ $key }->{ $key2 }->[ 1 ] }; ++$i ) {
			print { $fh_closest_gene } $closest_genes_of->{ $key }->{ $key2 }->[ 1 ][ $i ], "\t";
		}
		print { $fh_closest_gene } "\nGenes downstream: \n";
		for ( my $i = 0; $i < @{ $closest_genes_of->{ $key }->{ $key2 }->[ 2 ] }; ++$i ) {
			print { $fh_closest_gene } $closest_genes_of->{ $key }->{ $key2 }->[ 2 ][ $i ], "\t";
		}
		print { $fh_closest_gene } "\nGenes in cluster: \n";
		for ( my $i = 0; $i < @{ $closest_genes_of->{ $key }->{ $key2 }->[ 3 ] }; ++$i ) {
			print { $fh_closest_gene } $closest_genes_of->{ $key }->{ $key2 }->[ 3 ][ $i ], "\t";
		}
		print { $fh_closest_gene } "\n";
	}
}
close $fh_closest_gene;

exit;

#####################Subroutine(s)##############

# adds species abbreviation to scaffolds
sub addSpecies {
	my $sequence_of = shift @_;
	my $species 	= shift @_;

	my %species_sequence_of;
	
	foreach my $key (sort keys $sequence_of ) {
		$key =~ m/>(\S*).*/;
		my $key2 = $species . "_" . $1;
		$species_sequence_of{ $key2 } = $sequence_of->{ $key };
	}
	return \%species_sequence_of;
}

# get the nt sequence of each CNE from the fasta file of the reference species
sub getCNEsequence {
	my $CNEs        		= shift @_;
	my $genome_file 		= shift @_;
	my $cne_sequence_output = shift @_;
	my $species 			= shift @_;
	open my $output_fh, '>>', $cne_sequence_output or die "Couldn't open \"$cne_sequence_output\": $!\n";
	my $sequence_of = &readFasta( $genome_file );
	$sequence_of    = &addSpecies( $sequence_of, $species );
	my %cne_sequence_of;
	foreach my $key ( keys $CNEs ) {
		for ( my $i = 0; $i < @{ $CNEs->{ $key } }; ++$i ) {
			my $length = $CNEs->{ $key }[ $i ]{ 'end' } - $CNEs->{ $key }[ $i ]{ 'start' } + 1;
			my $sequence = substr( $sequence_of->{ $key }, $CNEs->{ $key }[ $i ]{ 'start' }, $length );
			print { $output_fh} ">", $key, "_", $CNEs->{ $key }[ $i ]{ 'start' }, "_", $CNEs->{ $key }[ $i ]{ 'end' }, "\n", $sequence, "\n";
		}
	}
	close $output_fh;
}

# get the closest genes of a CNE cluster using annotation file
sub getclosestgenes {
	my $CNE_cluster  = shift @_;
	my $gene_info_of = shift @_;
	my $species 	 = shift @_;

	my %CNE_cluster_genes;
	# output file
	my $file_cluster = "/home/tanja/holo_genomes_results/CNE/cne_in_cluster_" . $species. ".txt";
	open my $fh_cluster, '>', $file_cluster or die "Couldn't open file \"$file_cluster\": $!\n";

	print { $fh_cluster } "Scaffold\tCluster_start\tCluster_stop\tNo_of_CNEs_in_cluster\n";

	# go through each CNE cluster
	foreach my $key ( sort keys $CNE_cluster ) {
		# iterate through each CNE in a cluster
		foreach my $key2 ( sort keys $CNE_cluster->{ $key } ) {
			my $cluster_start;
			my $cluster_end;
			my $first = 1;

			# get cluster start and stop coordinates
			for ( my $i = 0; $i < @{ $CNE_cluster->{ $key }->{ $key2 } }; ++$i ) {
				if ( $first ) {
					$cluster_start = $CNE_cluster->{ $key }->{ $key2 }[ $i ]{ 'start' };
					$cluster_end   = $CNE_cluster->{ $key }->{ $key2 }[ $i ]{ 'end' };
					$first = 0;
				}
				else {
					if ( $CNE_cluster->{ $key }->{ $key2 }[ $i ]{ 'start' } < $cluster_start ) {
						$cluster_start = $CNE_cluster->{ $key }->{ $key2 }[ $i ]{ 'start' };
					}
					elsif ( $CNE_cluster->{ $key }->{ $key2 }[ $i ]{ 'end' } > $cluster_end ) {
						$cluster_end   = $CNE_cluster->{ $key }->{ $key2 }[ $i ]{ 'end' };
					}
				}
			}
			# print cluster info in output file
			print { $fh_cluster } ">", $key, "\t", $cluster_start, "\t", $cluster_end, "\t", scalar @{ $CNE_cluster->{ $key }->{ $key2 } }, "\n";
			for ( my $i = 0; $i < @{ $CNE_cluster->{ $key }->{ $key2 } }; ++$i ) {
				print{ $fh_cluster } $CNE_cluster->{ $key }->{ $key2 }[ $i ]{ 'start' }, "\t", $CNE_cluster->{ $key }->{ $key2 }[ $i ]{ 'end' }, "\n";
			}
			# add cluster info to hash
			$CNE_cluster_genes{ $key }{ $key2 }[ 0 ]{ 'start' } = $cluster_start;
			$CNE_cluster_genes{ $key }{ $key2 }[ 0 ]{ 'end' } = $cluster_end;
			$CNE_cluster_genes{ $key }{ $key2 }[ 0 ]{ 'cne_no' } = scalar @{ $CNE_cluster->{ $key }->{ $key2 } };

			my @genes_upstream;
			my @genes_downstream;
			my @genes_in_cluster;
			# add corresponding genes of cluster to hash
			foreach my $key3 ( sort keys $gene_info_of ) {
				next if $key =~ m/^TRNA/i;
				next if $key =~ m/^Mir/i;
				if ( $key eq $gene_info_of->{ $key3 }[ 0 ]{ 'scaffold' } ) {
					if ( $gene_info_of->{ $key3 }[ 0 ]{ 'end' } < $cluster_start && ( $cluster_start - $gene_info_of->{ $key3 }[ 0 ]{ 'end' } ) < 500000 ) {
						push @genes_upstream, $key3;
					}
					elsif ( ( $gene_info_of->{ $key3 }[ 0 ]{ 'start' } >= $cluster_start && $gene_info_of->{ $key3 }[ 0 ]{ 'start' } < $cluster_end ) or ( $gene_info_of->{ $key3 }[ 0 ]{ 'end' } > $cluster_start && $gene_info_of->{ $key3 }[ 0 ]{ 'end' } <= $cluster_end ) ) {
						push @genes_in_cluster, $key3;
					}
					elsif ( $gene_info_of->{ $key3 }[ 0 ]{ 'start' } > $cluster_end && ( $gene_info_of->{ $key3 }[ 0 ]{ 'start' } - $cluster_end ) < 500000 ) {
						push @genes_downstream, $key3;
					}
				}
			}
			# add na if no gnes could be found next to/in a cluster
			if ( !@genes_upstream ) {
				push @genes_upstream, 'na';
			}
			if ( !@genes_downstream ) {
				push @genes_downstream, 'na';
			}
			if ( !@genes_in_cluster ) {
				push @genes_in_cluster, 'na';
			}
			# add lists of genes to hash
			push @{ $CNE_cluster_genes{ $key }{ $key2 } }, \@genes_upstream;
			push @{ $CNE_cluster_genes{ $key }{ $key2 } }, \@genes_downstream;
			push @{ $CNE_cluster_genes{ $key }{ $key2 } }, \@genes_in_cluster;
		}
	}
	close $fh_cluster;
	return \%CNE_cluster_genes;
}

# find all CNEs that belong to the same cluster; same cluster = < 20,000 nt distance between two
sub getCNEcluster {
	my $CNEs_first  = shift @_;
	my $CNEs_second = shift @_;
	my $CNEs_third  = shift @_;
	
	my %cluster_of_CNEs;
	# to through each CNE from the CNEr output files
	# key is the scaffold
	foreach my $key ( sort keys $CNEs_first ) {
		print $key, "\n";
		my $count = 0;
		if ( exists $cluster_of_CNEs{ $key } ) {
			foreach my $key3 ( sort keys $cluster_of_CNEs{ $key } ) {
				if ( $count <= $key3 ) {
					$count = $key3 + 1;
				}
			}
		}
		# check if cluster exists for this scaffold already
		for ( my $i = 0; $i < @{ $CNEs_first->{ $key } }; ++$i ) {
			# if not add CNE to the hash using count as key
			if ( !exists $cluster_of_CNEs{ $key } ) {
				push @{ $cluster_of_CNEs{ $key }{ $count } }, { 'start' => $CNEs_first->{ $key }[ $i ]{ 'start' }, 'end' => $CNEs_first->{ $key }[ $i ]{ 'end' } };
				++$count;
			}
			# if it exists check for cluster the CNE could belong to
			else {
				# loop through all count IDs
				my $key_found;
				my $cluster_found = 0;
				# go through existing cluster
				foreach my $key2 ( sort keys $cluster_of_CNEs{ $key } ) {
					# go through CNEs in cluster
					for ( my $j = 0; $j < @{ $cluster_of_CNEs{ $key }{ $key2 } }; ++$j ) {
						my $end;
						my $start;
						if ( $CNEs_first->{ $key }[ $i ]{ 'end' } < $cluster_of_CNEs{ $key }{ $key2 }[ $j ]{ 'start' } ) {
							$end = $CNEs_first->{ $key }[ $i ]{ 'end' };
							$start = $cluster_of_CNEs{ $key }{ $key2 }[ $j ]{ 'start' };
						}
						else {
							$end = $cluster_of_CNEs{ $key }{ $key2 }[ $j ]{ 'end' };
							$start = $CNEs_first->{ $key }[ $i ]{ 'start' };
						}
						if ( abs ( $start - $end ) <= 20000 ) {
							$cluster_found = 1; 
							$key_found = $key2;
							last;
						}
					}	
				}
				# add to existing cluster if found
				if ( $cluster_found ) {
					push @{ $cluster_of_CNEs{ $key }{ $key_found } }, { 'start' => $CNEs_first->{ $key }[ $i ]{ 'start' }, 'end' => $CNEs_first->{ $key }[ $i ]{ 'end' } };
				}
				# create new cluster if not found
				else {
					push @{ $cluster_of_CNEs{ $key }{ $count } }, { 'start' => $CNEs_first->{ $key }[ $i ]{ 'start' }, 'end' => $CNEs_first->{ $key }[ $i ]{ 'end' } };
					++$count;
				}
			}
		}
	}
	#repeat for second file
	foreach my $key ( sort keys $CNEs_second ) {
		print "2 ", $key, "\n";
		my $count = 0;
		if ( exists $cluster_of_CNEs{ $key } ) {
			foreach my $key3 ( sort keys $cluster_of_CNEs{ $key } ) {
				if ( $count <= $key3 ) {
					$count = $key3 + 1;
				}
			}
		}
		for ( my $i = 0; $i < @{ $CNEs_second->{ $key } }; ++$i ) {
			if ( !exists $cluster_of_CNEs{ $key } ) {
				push @{ $cluster_of_CNEs{ $key }{ $count } }, { 'start' => $CNEs_second->{ $key }[ $i ]{ 'start' }, 'end' => $CNEs_second->{ $key }[ $i ]{ 'end' } };
				++$count;
			}
			else {
				# loop through all count IDs
				my $key_found;
				my $cluster_found = 0;
				foreach my $key2 ( sort keys $cluster_of_CNEs{ $key } ) {
					
					for ( my $j = 0; $j < @{ $cluster_of_CNEs{ $key }{ $key2 } }; ++$j ) {
						my $end;
						my $start;
						if ( $CNEs_second->{ $key }[ $i ]{ 'end' } < $cluster_of_CNEs{ $key }{ $key2 }[ $j ]{ 'start' } ) {
							$end = $CNEs_second->{ $key }[ $i ]{ 'end' };
							$start = $cluster_of_CNEs{ $key }{ $key2 }[ $j ]{ 'start' };
						}
						else {
							$end = $cluster_of_CNEs{ $key }{ $key2 }[ $j ]{ 'end' };
							$start = $CNEs_second->{ $key }[ $i ]{ 'start' };
						}
						if ( abs ( $start - $end ) <= 20000 ) {
							$cluster_found = 1; 
							$key_found = $key2;
							last;
						}
					}	
				}
				if ( $cluster_found ) {
					push @{ $cluster_of_CNEs{ $key }{ $key_found } }, { 'start' => $CNEs_second->{ $key }[ $i ]{ 'start' }, 'end' => $CNEs_second->{ $key }[ $i ]{ 'end' } };
				}
				else {
					push @{ $cluster_of_CNEs{ $key }{ $count } }, { 'start' => $CNEs_second->{ $key }[ $i ]{ 'start' }, 'end' => $CNEs_second->{ $key }[ $i ]{ 'end' } };
					++$count;
				}
				
			}
		}
	}
	# repeat for third file
	foreach my $key ( sort keys $CNEs_third ) {
		print "3 ", $key, "\n";
		my $count = 0;
		if ( exists $cluster_of_CNEs{ $key } ) {
			foreach my $key3 ( sort keys $cluster_of_CNEs{ $key } ) {
				if ( $count <= $key3 ) {
					$count = $key3 + 1;
				}
			}
		}
		for ( my $i = 0; $i < @{ $CNEs_third->{ $key } }; ++$i ) {
			if ( !exists $cluster_of_CNEs{ $key } ) {
				push @{ $cluster_of_CNEs{ $key }{ $count } }, { 'start' => $CNEs_third->{ $key }[ $i ]{ 'start' }, 'end' => $CNEs_third->{ $key }[ $i ]{ 'end' } };
				++$count;
			}
			else {
				# loop through all count IDs
				my $key_found;
				my $cluster_found = 0;
				foreach my $key2 ( sort keys $cluster_of_CNEs{ $key } ) {
					for ( my $j = 0; $j < @{ $cluster_of_CNEs{ $key }{ $key2 } }; ++$j ) {
						my $end;
						my $start;
						if ( $CNEs_third->{ $key }[ $i ]{ 'end' } < $cluster_of_CNEs{ $key }{ $key2 }[ $j ]{ 'start' } ) {
							$end = $CNEs_third->{ $key }[ $i ]{ 'end' };
							$start = $cluster_of_CNEs{ $key }{ $key2 }[ $j ]{ 'start' };
						}
						else {
							$end = $cluster_of_CNEs{ $key }{ $key2 }[ $j ]{ 'end' };
							$start = $CNEs_third->{ $key }[ $i ]{ 'start' };
						}
						if ( abs ( $start - $end ) <= 20000 ) {
							$cluster_found = 1; 
							$key_found = $key2;
							last;
						}
					}
				}	
				if ( $cluster_found == 1 ) {
					push @{ $cluster_of_CNEs{ $key }{ $key_found } }, { 'start' => $CNEs_third->{ $key }[ $i ]{ 'start' }, 'end' => $CNEs_third->{ $key }[ $i ]{ 'end' } };
				}
				else {
					push @{ $cluster_of_CNEs{ $key }{ $count } }, { 'start' => $CNEs_third->{ $key }[ $i ]{ 'start' }, 'end' => $CNEs_third->{ $key }[ $i ]{ 'end' } };
					++$count;
				}
			}
		}
	}
	return \%cluster_of_CNEs;
}

# get the nt length of each scaffold of the reference species
sub getScaffoldlength {
	my $genome_file = shift @_;
	my $species 	= shift @_;
	my $sequence_of = &readFasta( $genome_file );
	$sequence_of 	= &addSpecies( $sequence_of, $species );

	my %scaffold_length;
	foreach my $key ( keys $sequence_of ){
		my $length = length $sequence_of->{ $key };
		$scaffold_length{ $key } = $length;
	}
	return \%scaffold_length;
}

# remove redundancy in CNEs
# overlapping CNEs will be combined, start/stop will be changed
sub getUniqueCNEs {
	my $CNE_unique = shift @_;
	my $CNE_list   = shift @_;

	# iterate through all CNEs
	foreach my $key ( keys $CNE_list ) {
		# check if they exist in the unique list
		if ( !exists $CNE_unique->{ $key } ) {
			# if scaffold doesn't exist yet push all CNE to it
			# test for overlapping
			my $first = 1;
			for ( my $i = 0; $i < @{ $CNE_list->{ $key } }; ++$i ) {
				# always push first CNE to list
				if ( $first ) {
					push @{ $CNE_unique->{ $key } }, { 'start' => $CNE_list->{ $key }[ $i ]{ 'start' }, 'end' => $CNE_list->{ $key }[ $i ]{ 'end' } };
					$first = 0;
				}
				# check each following one for overlap
				else {
					my $found = 0;
					for ( my $j = 0; $j < @{ $CNE_unique->{ $key } }; ++$j ) {
						# test for inrange of start/stop of CNE
						# in both directions in case new CNE is longer in one direction
						my $range1 = Number::Range->new( $CNE_unique->{ $key }[ $j ]{ 'start' }..$CNE_unique->{ $key }[ $j ]{ 'end' } );
						my $range2 = Number::Range->new( $CNE_list->{ $key }[ $i ]{ 'start' }..$CNE_list->{ $key }[ $i ]{ 'end' } );
						my @inrange1 = $range1->inrange( $CNE_list->{ $key }[ $i ]{ 'start' }, $CNE_list->{ $key }[ $i ]{ 'end' } );
						my @inrange2 = $range2->inrange( $CNE_unique->{ $key }[ $j ]{ 'start' }, $CNE_unique->{ $key }[ $j ]{ 'end' } );
						# adapt borders of CNE if inrange is 1 and 0 at the same time
						if ( $inrange1[ 0 ] or $inrange1[ 1 ] or $inrange2[ 0 ] or $inrange2[ 1 ] ) {
							if ( $inrange1[ 0 ] == 0 && $inrange2[ 0 ] == 1 ) {
								$CNE_unique->{ $key }[ $j ]{ 'start' } = $CNE_list->{ $key }[ $i ]{ 'start' };
								$found = 1;
							}
							if ( $inrange1[ 1 ] == 0 && $inrange2[ 1 ] == 1 ) {
								$CNE_unique->{ $key }[ $j ]{ 'end' } = $CNE_list->{ $key }[ $i ]{ 'end' };
								$found = 1;
							}
						}
					}
					# if no overlapping CNE was found push it to the list
					if ( !$found ) {
						push @{ $CNE_unique->{ $key } }, { 'start' => $CNE_list->{ $key }[ $i ]{ 'start' }, 'end' => $CNE_list->{ $key }[ $i ]{ 'end' } };
					}
				}
			}
		}
		# if scaffold already exists in list check for overlap, change start/stop in case or add it
		else {
			for ( my $i = 0; $i < @{ $CNE_list->{ $key } }; ++$i ) {
				my $found = 0;
				for ( my $j = 0; $j < @{ $CNE_unique->{ $key } }; ++$j ){
					my $range1 = Number::Range->new( $CNE_unique->{ $key }[ $j ]{ 'start' }..$CNE_unique->{ $key }[ $j ]{ 'end' } );
					my $range2 = Number::Range->new( $CNE_list->{ $key }[ $i ]{ 'start' }..$CNE_list->{ $key }[ $i ]{ 'end' } );
					my @inrange1 = $range1->inrange( $CNE_list->{ $key }[ $i ]{ 'start' }, $CNE_list->{ $key }[ $i ]{ 'end' } );
					my @inrange2 = $range2->inrange( $CNE_unique->{ $key }[ $j ]{ 'start' }, $CNE_unique->{ $key }[ $j ]{ 'end' } );		
					if ( $inrange1[ 0 ] or $inrange1[ 1 ] or $inrange2[ 0 ] or $inrange2[ 1 ] ) {
						if ( $inrange1[ 0 ] == 0 && $inrange2[ 0 ] == 1 ) {
								$CNE_unique->{ $key }[ $j ]{ 'start' } = $CNE_list->{ $key }[ $i ]{ 'start' };
								$found = 1;
						}
						if ( $inrange1[ 1 ] == 0 && $inrange2[ 1 ] == 1 ) {
								$CNE_unique->{ $key }[ $j ]{ 'end' } = $CNE_list->{ $key }[ $i ]{ 'end' };
								$found = 1;
						}
					}
				}
				if ( !$found ) {
					push @{ $CNE_unique->{ $key } }, { 'start' => $CNE_list->{ $key }[ $i ]{ 'start' }, 'end' => $CNE_list->{ $key }[ $i ]{ 'end' } };
				}
			}
		}
	}
	return $CNE_unique;
}

# get gene names with scaffold, start, stop, and strand information from gff of reference species
sub readAnnotation {
	my $annotation_file = shift @_;
	my $species 		= shift @_;
	open my $fh_annotation, '<', $annotation_file or die "Couldn't open file: \"$annotation_file\": $!\n";
	my %gene_info_of;
	while ( my $line = <$fh_annotation> ) {
		chomp $line;
		if ( $line =~ m/\tgene\t/ or $line =~ m/\ttranscript\t/ ){
			#Spalte 1 = Scaffold, Spalte 4 = start, Spalte 5 = stop, Spalte 7 = Strang, Spalte 9 = Genname + anderes (Name oder GeneID als Identifier?)
			my @parts    = split( "\t", $line );
			$parts[ 0 ] =~ m/[a-z]*_?(.*)/i;
			my $scaffold = $species . "_" . $parts[ 0 ];
			#my $scaffold = $species . "_" . $1;
			my $start    = $parts[ 3 ];
			my $end      = $parts[ 4 ];
			my $strand   = $parts[ 6 ];
			# Nasonia und Apis
			#$parts[ 8 ]  =~ m/name=([^;]+);/i;
			# Athalia und Orussus
			$parts[ 8 ] =~ m/ID=([^;]+);/i;
			my $genename = $1;
			push @{ $gene_info_of{ $genename } }, { 'scaffold' => $scaffold, 'start' => $start, 'end' => $end, 'strand' => $strand };
		}
	}
	close $fh_annotation;
	return \%gene_info_of;
}

# get all CNEs from CNEr output file
sub readCNEs {
	my $CNE_file = shift @_;
	open my $fh_CNEs, '<', $CNE_file or die "Couldn't open file: \"$CNE_file\": $!\n";
	my %CNEs;
	while ( my $line = <$fh_CNEs> ) {
		next if $line =~ m/^track name/;
		chomp $line;
		my @parts = split( "\t", $line );
		push @{ $CNEs { $parts[ 0 ] } }, { 'start' => $parts[ 1 ], 'end' => $parts[ 2 ] };
	}
	
	close $fh_CNEs;
	return \%CNEs;
}

# read fasta file into hash, header as key
sub readFasta {
	my $genome_file = shift @_;
	open my $fh_genome, '<', $genome_file or die "Couldn't open file: \"$genome_file\": $!\n";

	my %sequence_of;
	my $first_header = 1;
	my $header;
	my $sequence;

	while ( my $line = <$fh_genome> ) {
		chomp $line;
		if ( $line =~ m/^>/ ) {
			if ( $first_header ) {
				$header = $line;
				$first_header = 0;
				next;
			}
			$sequence_of{ $header } = $sequence;
			$header = $line;
			$sequence = '';
		}
		else {
			$sequence .= $line;
		}
	}
	close $fh_genome;
	return \%sequence_of;
}