#!/usr/bin/perl

use strict;
use warnings;

# Reference species
my $species = 'oabi';

# Input file
my $closest_gene_file = "/home/tanja/holo_genomes_results/CNE/cne_closest_gene_unique_" . $species . ".tab";
open my $fh_closest_gene, '<', $closest_gene_file or die "Couldn't open file \"$closest_gene_file\": $!\n";

my %gene_count;
my %gene_info_of;

my $key;
# count how often one gene occurs
while ( my $line = <$fh_closest_gene> ) {
	next if $line =~ m/^#/;
	chomp $line;
	if ( $line =~ m/^>/ ) {
		$key = $line;
	}
	else {
		push @{ $gene_info_of{ $key } }, $line;
	}
	if ( $line !~ m/>|Genes|na/ ) {
		my @parts = split ( "\t", $line );
		foreach my $part ( @parts ) {
			$gene_count{ $part } += 1;
		}
	}
}

close $fh_closest_gene;

# Output file, list of genes occuring in neighbourhood of CNE cluster with count
my $gene_list_file = "/home/tanja/holo_genomes_results/CNE/gene_list_cne_" . $species . ".tab";
open my $fh_gene_list, '>', $gene_list_file or die "Couldn't open file \"$gene_list_file\": $!\n";

print { $fh_gene_list } "Gene_ID\tTotal_count\n";

foreach my $key ( sort keys %gene_count ) {
	print { $fh_gene_list } $key, "\t", $gene_count{ $key }, "\n";
}
close $fh_gene_list;

# output file, list of CNE clusters with count of genes occurring upstream, downstream, in cluster
my $cluster_gene_count_file = "/home/tanja/holo_genomes_results/CNE/gene_in_cluster_count_allcne_" . $species . ".csv";
open my $fh_cluster_gene, '>', $cluster_gene_count_file or die "Couldn't open file \"$cluster_gene_count_file\":$!\n";
# output file, list of CNE clusters with count of genes occurring upstream, downstream, in cluster with 500 kb distance to scaffold end
my $cluster_gene_count_500_file = "/home/tanja/holo_genomes_results/CNE/gene_in_cluster_count_500cne_" . $species . ".csv";
open my $fh_cluster_gene_500, '>', $cluster_gene_count_500_file or die "Couldn't open file \"$cluster_gene_count_500_file\": $!\n";

print { $fh_cluster_gene } "# Scaffold\tCluster_start\tCluster_end\tCNE_number_in_cluster\tDistance_to_Scaffold_end\tGenes_upstream\tGenes_downstream\tGenes_in_cluster\n";
print { $fh_cluster_gene_500 } "# Scaffold\tCluster_start\tCluster_end\tCNE_number_in_cluster\tDistance_to_Scaffold_end\tGenes_upstream\tGenes_downstream\tGenes_in_cluster\n";

foreach my $key ( sort keys %gene_info_of ) {
	my $count_downstream = 0;
	my $count_upstream = 0;
	my $count_in_cluster = 0;
	my @parts = split ( "\t", $key );
	if ( $gene_info_of{ $key }[ 1 ] !~ m/^na/ ) {
		my @parts_downstream = split ( "\t", $gene_info_of{ $key }[ 1 ] );
		$count_downstream = scalar @parts_downstream;
	}
	if ( $gene_info_of{ $key }[ 3 ] !~ m/^na/ ) {
		my @parts_upstream = split ( "\t", $gene_info_of{ $key }[ 3 ] );
		$count_upstream = scalar @parts_upstream;
	}
	if ( $gene_info_of{ $key }[ 5 ] !~ m/^na/ ) {
		my @parts_cluster = split ( "\t", $gene_info_of{ $key }[ 5 ] );
		$count_in_cluster = scalar @parts_cluster;
	}
	print { $fh_cluster_gene } $key, "\t", $count_downstream, "\t", $count_upstream, "\t", $count_in_cluster, "\n";
	if ( $parts[ 1 ] >= 500000 && $parts[ 4 ] >= 500000 ) {
		print { $fh_cluster_gene_500 } $key, "\t", $count_downstream, "\t", $count_upstream, "\t", $count_in_cluster, "\n";
	}
}
close $fh_cluster_gene;
close $fh_cluster_gene_500;