#!/usr/bin/perl

use strict;
use warnings;
use Number::Range;

# Reference species
my $species = 'aros';

# file with all genes (upstream, downstream, in cluster) next to a clsuter with distance < 500 kb 
my $gene_file_in = "/home/tanja/holo_genomes_results/CNE/cne_closest_gene_unique_" . $species . ".tab";

open my $fh_gene_in, '<', $gene_file_in or die "Couldn't open file \"$gene_file_in\": $!\n";

# gff file of reference species
#my $gff_file = "/home/tanja/holo_genomes/Amel_4.5_top_level_mod_cner.gff3";
my $gff_file = "/home/tanja/holo_genomes/ATHROS_OGS_with_lnc_sorted.gff";
#my $gff_file = "/home/tanja/holo_genomes/ref_Nvit_2.1_top_level_genebank.gff3";
#my $gff_file = "/home/tanja/holo_genomes_results/ORUABI_OGSv1-0_withlncrna_sorted.gff3";

open my $fh_gff, '<', $gff_file or die "Couldn't open file \"$gff_file\": $!\n";

# create hash containing all genes next to a cluster
my %genes_of;
my $key;
while ( my $line = <$fh_gene_in> ) {
	chomp $line;
	next if $line =~ m/^#/;
	#next if $line =~ m/^na/;
	next if $line =~ m/^Genes/;
	if ( $line =~ m/^>/ ) {
		$key = $line;
	}
	else {
		my @parts = split ( "\t", $line );
		foreach my $part ( @parts ) {
			push @{ $genes_of { $key } }, $part;
		}
	}
}

my %gene_info_of;

# create a hash containing all genes from gff with position and strand info
while ( my $line = <$fh_gff> ) {
	chomp $line;
	if ( $line =~ m/\tgene\t/ or $line =~ m/\ttranscript/ ){
		#Spalte 1 = Scaffold, Spalte 4 = start, Spalte 5 = stop, Spalte 7 = Strang, Spalte 9 = Genname + anderes (Name oder GeneID als Identifier?)
		my @parts    = split( "\t", $line );
		$parts[ 0 ] =~ m/[a-z]*_?(.*)/i;
		my $scaffold = ">" . $species . "_" . $parts[ 0 ];
#		my $scaffold = $species . "_" . $parts[ 0 ];
		my $start    = $parts[ 3 ];
		my $end      = $parts[ 4 ];
		my $strand   = $parts[ 6 ];
		# Nasonia und Apis
		#$parts[ 8 ]  =~ m/name=([^;]+);/i;
		# Athalia und Orussus
		$parts[ 8 ]  =~ m/ID=([^;]+);/i;
		my $genename = $1;
		push @{ $gene_info_of{ $genename } }, { 'scaffold' => $scaffold, 'start' => $start, 'end' => $end, 'strand' => $strand };
	}
}

close $fh_gene_in;
close $fh_gff;

my %closest_gene_of_upstream;
my %closest_gene_of_downstream;

# find closest gene of cluster
foreach my $key ( sort keys %genes_of ) {
	# 0 = >scaffold, 1 = start, 2 = end, 3 = gene count, 4 = distance to end of scaffold
	print $key, "\n";
	my @parts = split ( "\t", $key );
	foreach my $gene ( @{ $genes_of{ $key } } ) {
		# skip miRNAs and tRNAs
		next if $gene =~ m/^mir/i;
		next if $gene =~ m/^trna/i;
		next if $gene =~ m/bantam/i;
		print "gene: ", $gene, "\n";
		# get gene info
		my $gene_start = $gene_info_of{ $gene }[ 0 ]{ 'start' };
		my $gene_end   = $gene_info_of{ $gene }[ 0 ]{ 'end' };
		my $strand	   = $gene_info_of{ $gene }[ 0 ]{ 'strand' };
		# check if upstream gene is closest to the cluster
		if ( $gene ne 'na' ) {
			if ( $strand eq '-' ) {
				if ( $gene_end < $parts[ 1 ] && abs( $gene_end - $parts[ 1 ] ) < 500000 ) {
					# create new key if no gene exists for this cluster
					if ( !exists $closest_gene_of_upstream{ $key } ) {
						print "!exist upstream\n";
						$closest_gene_of_upstream{ $key } = $gene;
					}
					# if it exists check distance 
					else {
						print "exist upstream\n";
						my $key_old = $closest_gene_of_upstream{ $key };
						my $dist_old = abs( $parts[ 1 ] - $gene_info_of{ $key_old }[ 0 ]{ 'end' } );
						if ( $gene_end < $parts[ 1 ] ) {
							my $dist_new = abs( $parts[ 1 ] - $gene_end );
							print $dist_old, "\t", $dist_new, "\n";
							print $closest_gene_of_upstream{ $key }, "\t";
							$closest_gene_of_upstream{ $key } = $gene if $dist_new < $dist_old;
							print $closest_gene_of_upstream{ $key }, "\n";
						}
					}
				}
			}
			# check if downstream gene is closest to the cluster
			else {
				if ( $gene_start > $parts[2 ] && abs ( $gene_start - $parts[ 2 ] ) < 500000 ) {
					if ( !exists $closest_gene_of_downstream{ $key } ) {
						print "!exist downstream\n";
						$closest_gene_of_downstream{ $key } = $gene;
					}
					else {
						print "exist downstream\n";
						my $key_old = $closest_gene_of_downstream{ $key };
						my $dist_old = abs( $gene_info_of{ $key_old }[ 0 ]{ 'start' } - $parts[ 2 ] );
						if ( $gene_start > $parts[ 2 ] ) {
							my $dist_new = abs( $gene_start - $parts[ 2 ] );
							print $dist_old, "\t", $dist_new, "\n";
							print $closest_gene_of_downstream{ $key }, "\t";
							$closest_gene_of_downstream{ $key } = $gene if $dist_new < $dist_old;
							print $closest_gene_of_downstream{ $key }, "\n";
						}
					}
				}
			}
		}
	}
	# print na if no gene in right orientation is found
	if ( !exists $closest_gene_of_upstream{ $key } ) {
		print "not exists upstream\n";
		$closest_gene_of_upstream{ $key } = 'na';
	}
	if ( !exists $closest_gene_of_downstream{ $key } ) {
		print "not exists downstream\n";
		$closest_gene_of_downstream{ $key } = 'na';
	}
	print $closest_gene_of_upstream{ $key }, "\n", $closest_gene_of_downstream{ $key }, "\n";
}

# output files
# file with closest genes on both sides of cluster in cis direction 
my $file_out = "/home/tanja/holo_genomes_results/CNE/cne_closest_two_genes_all_" . $species . ".tab";
open my $fh_out, '>', $file_out or die "Couldn't open file \"$file_out\": $!\n";

# file with closest genes on both sides in trans direction
my $file_out_trans = "/home/tanja/holo_genomes_results/CNE/cne_closest_two_genes_trans_all_" . $species . ".tab";
open my $fh_out_trans, '>', $file_out_trans or die "Couldn't open file \"$file_out\": $!\n";

print { $fh_out } "# Scaffold\tCluster_start\tCluster_end\tDistance_to_scaffold_end\tClosest_gene_upstream\tDistance_to_gene_upstream\tClosest_gene_downstream\tDistance_to_gene_downstream\n";
print { $fh_out_trans } "# Scaffold\tCluster_start\tCluster_end\tDistance_to_scaffold_end\tClosest_gene_upstream_trans\tDistance_to_gene_upstream_trans\tClosest_gene_downstream_trans\tDistance_to_gene_downstream_trans\n";

# if closest gene to cluster is in trans direction print 'na'
foreach my $key ( sort keys %closest_gene_of_upstream ) {
	my @parts 	   = split ( "\t", $key );
	print $key, "\n";
#	next if $parts[ 1 ] < 500000;
#	next if $parts[ 4 ] < 500000;
	my $gene_upstream = $closest_gene_of_upstream{ $key };
	print $gene_upstream, "\n";
	#exit;
	my $gene_end_upstream;
	my $dist_upstream;
	my $gene_upstream_trans = 'na';
	my $gene_end_upstream_trans = '-';
	my $dist_upstream_trans = '-';
	if ( $gene_upstream eq 'na' ) {
		$gene_end_upstream = '-';
		$dist_upstream = '-';
		$gene_end_upstream_trans = '-';
		$dist_upstream_trans = '-';
	}
	else {
		$gene_end_upstream = $gene_info_of{ $gene_upstream }[ 0 ]{ 'end' };
		$dist_upstream 	   = abs( $parts[ 1 ] - $gene_end_upstream );
		foreach my $gene ( @{ $genes_of{ $key } } ) {
			next if $gene =~ m/^mir/i;
			next if $gene =~ m/^trna/i;
			next if $gene =~ m/bantam/i;
			#next if $gene =~ m/lncRNA/i;
			my $gene_start = $gene_info_of{ $gene }[ 0 ]{ 'start' };
			my $gene_end   = $gene_info_of{ $gene }[ 0 ]{ 'end' };
			my $strand	   = $gene_info_of{ $gene }[ 0 ]{ 'strand' };
			if ( $strand eq '+' && $gene_start < $parts[ 1 ] && $gene_start > $gene_end_upstream) {
				$gene_upstream_trans = $gene;
				$gene_end_upstream_trans = $gene_info_of{ $gene }[ 0 ]{ 'end' };
				$dist_upstream_trans = abs( $parts[ 1 ] - $gene_end_upstream_trans );
				$gene_upstream = 'na';
				$gene_end_upstream = '-';
				$dist_upstream = '-';
				last;
			}	
		}
	}
	my $gene_downstream = $closest_gene_of_downstream{ $key };
	my $gene_start_downstream;
	my $dist_downstream;
	my $gene_downstream_trans = 'na';
	my $gene_start_downstream_trans = '-';
	my $dist_downstream_trans = '-';
	if ( $gene_downstream eq 'na' ) {
		$gene_start_downstream = '-';
		$dist_downstream = '-';
	}
	else {
		$gene_start_downstream = $gene_info_of{ $gene_downstream }[ 0 ]{ 'start' };
		$dist_downstream 	   = abs( $gene_start_downstream - $parts[ 2 ] );
		foreach my $gene ( @{ $genes_of{ $key } } ) {
			next if $gene =~ m/^mir/i;
			next if $gene =~ m/^trna/i;
			next if $gene =~ m/bantam/i;
			my $gene_start = $gene_info_of{ $gene }[ 0 ]{ 'start' };
			my $gene_end   = $gene_info_of{ $gene }[ 0 ]{ 'end' };
			my $strand	   = $gene_info_of{ $gene }[ 0 ]{ 'strand' };
			if ( $strand eq '-' && $gene_start > $parts[ 2 ] && $gene_start < $gene_start_downstream ) {
				$gene_downstream_trans = $gene;
				$gene_start_downstream_trans = $gene_info_of{ $gene }[ 0 ]{ 'start' };
				$dist_downstream_trans = abs( $gene_start_downstream_trans - $parts[ 2 ] );
				$gene_downstream = 'na';
				$gene_start_downstream = '-';
				$dist_downstream = '-';
				last;
			}
		}
	}
	print $gene_upstream, "\t", $gene_downstream, "\n";
	
	print { $fh_out } $parts[ 0 ], "\t", $parts[ 1 ], "\t", $parts[ 2 ], "\t", $parts[ 4 ], "\t", $gene_upstream, "\t", $dist_upstream, "\t", $gene_downstream, "\t", $dist_downstream, "\n";
	print { $fh_out_trans } $parts[ 0 ], "\t", $parts[ 1 ], "\t", $parts[ 2 ], "\t", $parts[ 4 ], "\t", $gene_upstream_trans, "\t", $dist_upstream_trans, "\t", $gene_downstream_trans, "\t", $dist_downstream_trans, "\n";
}

close $fh_out;